@@ -3,50 +3,40 @@
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-import os, sys, subprocess
+import os
+import subprocess
+from .compat import str, sys_encoding
 
-def _wrapcall(*args, **kwargs):
+def _call(*args, **kwargs):
     try:
-        return subprocess.check_output(*args, **kwargs).decode().splitlines()
-    except AttributeError:
-        return _wrapcall_2_6(*args, **kwargs)
-    except subprocess.CalledProcessError:
-        return []
-
-def _wrapcall_2_6(*args, **kwargs):
-    try:
-        # no check_output in 2.6
-        if "stdout" in kwargs:
-            raise ValueError("stdout argument not allowed, it will be overridden.")
-        process = subprocess.Popen(stdout=subprocess.PIPE, *args, **kwargs)
-        output, unused_err = process.communicate()
-        retcode = process.poll()
-        if retcode:
-            cmd = kwargs.get("args")
-            if cmd is None:
-                cmd = args[0]
-            raise subprocess.CalledProcessError(retcode, cmd)
-        return output.decode().splitlines()
+        return subprocess.check_output(*args, **kwargs).decode(sys_encoding).splitlines()
     except subprocess.CalledProcessError:
         return []
 
 class ChoicesCompleter(object):
-    def __init__(self, choices=[]):
+    def __init__(self, choices):
         self.choices = choices
 
-    def __call__(self, prefix, **kwargs):
-        return (c for c in self.choices if c.startswith(prefix))
+    def _convert(self, choice):
+        if isinstance(choice, bytes):
+            choice = choice.decode(sys_encoding)
+        if not isinstance(choice, str):
+            choice = str(choice)
+        return choice
 
-def EnvironCompleter(prefix, **kwargs):
-    return (v for v in os.environ if v.startswith(prefix))
+    def __call__(self, **kwargs):
+        return (self._convert(c) for c in self.choices)
+
+EnvironCompleter = ChoicesCompleter(os.environ)
 
 class FilesCompleter(object):
     """
     File completer class, optionally takes a list of allowed extensions
     """
-    def __init__(self,allowednames=(),directories=True):
+
+    def __init__(self, allowednames=(), directories=True):
         # Fix if someone passes in a string instead of a list
-        if type(allowednames) is str:
+        if isinstance(allowednames, (str, bytes)):
             allowednames = [allowednames]
 
         self.allowednames = [x.lstrip("*").lstrip(".") for x in allowednames]
@@ -56,14 +46,14 @@ class FilesCompleter(object):
         completion = []
         if self.allowednames:
             if self.directories:
-                files = _wrapcall(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
-                completion += [ f + "/" for f in files]
+                files = _call(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
+                completion += [f + "/" for f in files]
             for x in self.allowednames:
-                completion += _wrapcall(["bash", "-c", "compgen -A file -X '!*.{0}' -- '{p}'".format(x,p=prefix)])
+                completion += _call(["bash", "-c", "compgen -A file -X '!*.{0}' -- '{p}'".format(x, p=prefix)])
         else:
-            completion += _wrapcall(["bash", "-c", "compgen -A file -- '{p}'".format(p=prefix)])
-            anticomp = _wrapcall(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
-            completion = list( set(completion) - set(anticomp))
+            completion += _call(["bash", "-c", "compgen -A file -- '{p}'".format(p=prefix)])
+            anticomp = _call(["bash", "-c", "compgen -A directory -- '{p}'".format(p=prefix)])
+            completion = list(set(completion) - set(anticomp))
 
             if self.directories:
                 completion += [f + "/" for f in anticomp]
@@ -88,7 +78,7 @@ class _FilteredFilesCompleter(object):
         try:
             names = os.listdir(target_dir or ".")
         except:
-            return # empty iterator
+            return  # empty iterator
         incomplete_part = os.path.basename(prefix)
         # Iterate on target_dir entries and filter on given predicate
         for name in names:
